# -*- coding: utf-8 -*- 
# @Time    : 03/22/2024 
# @Author  : nzooherd 
# @File    : __init__.py 
# @Software: Visual Studio Code

import functools
from logging import getLogger
from typing import Callable

from sqlalchemy import create_engine, select
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine

from component.handle_context import HandleContext
from config import Config

logger = getLogger()

class MySession:
    """
    """

    sync_engine = create_engine(Config.DATABASE_URL, pool_size=32, max_overflow=64)
    sessionfactory = async_sessionmaker(bind=create_async_engine(
        Config.DATABASE_URL, pool_size=32, max_overflow=64))
    handle_context = HandleContext()


    def __init__(self, mannul_tx=False):
        if s := MySession.handle_context.get_tx_session():
            self.session = s
            logger.debug(f"Use cache session: {id(self.session)}")
            self.use_cache = True
        else:
            self.session = MySession.sessionfactory()
            logger.debug(f"Make new session: {id(self.session)}")
            self.use_cache = False
        
        self.mannul_tx = mannul_tx
        self.s = None
        self.sb = None
    
    
    def __getattr__(self, item: str):
        return getattr(self.session, item)

    async def __aenter__(self):
        if self.mannul_tx:
            MySession.handle_context.set_tx_session(self.session)

        if not self.use_cache:
            self.s = await self.session.__aenter__()
            self.sb = self.s.begin()
            await self.sb.__aenter__()
        return self

    async def __aexit__(self, *args, **kwargs):
        if self.mannul_tx:
            MySession.handle_context.set_tx_session(None)
        
        if self.use_cache:
            await self.session.flush()

        if not self.use_cache:
            await self.sb.__aexit__(*args, **kwargs)
            await self.session.__aexit__(*args, **kwargs)

class BaseDao:

    @classmethod
    def session(cls) -> MySession:
        return MySession()

    @classmethod
    async def insert(cls, model):
        async with cls.session() as s:
            s.add(model)
            await s.flush()
            await s.refresh(model)
            s.expunge(model)
            return model
    
    @classmethod
    async def bulk_insert(cls, models):
        async with cls.session() as s:
            s.add_all(models)

    @classmethod
    async def select_by_id(cls, id):
        async with cls.session() as s:
            query = select(cls.model).where(cls.model.id == id)
            result = await s.execute(query)
            item = result.scalars().first()
            s.expunge(item)
            return item

    @classmethod
    async def select_all(cls, attrs=None):
        async with cls.session() as s:
            if not attrs:
                attrs = cls.model
            query = select(attrs)
            result = await s.execute(query)
            items = result.scalars().all()
            s.expunge_all()
            return items

def transaction(handler: Callable):

    @functools.wraps(handler)
    async def wrapper(*args, **kwargs):
        async with MySession(mannul_tx=True):
            return await handler(*args, **kwargs)

    return wrapper